#ifndef ARGSET_TRANSFORM_HH
#define ARGSET_TRANSFORM_HH

#include "paramlists.hh"
#include "argsets.hh"
#include "exprtype.hh"
#include "ctmax.hh"

namespace ArgSets
{
    template<typename Arg,typename CAS>
    const Arg& transformArg(const Arg& arg,const CAS&)
    {
	return arg;
    }
    
    template<unsigned i,typename CAS>
    typename GetASEntry<CAS,i>::Type& transformArg(const FuncImpl::ParamType<i>&,const CAS& cas)
    {
	return GetASEntry<CAS,i>::getVal(cas);
    }

    template<typename ArgSet,typename CAS>
    struct TransformArgSetHelper;
    
    template<typename CAS>
    struct TransformArgSetHelper<ArgSet0,CAS>;

    template<typename Arg,typename CAS>
    struct TransformArgSetResult
    {
	typedef const Arg& Type;
    };

    template<unsigned i,typename CAS>
    struct TransformArgSetResult<FuncImpl::ParamType<i>,CAS>
    {
	typedef typename GetASEntry<CAS,i>::Type& Type;
    };

    template<typename Arg,typename Base>
    struct ArgSetRefIfRef
    {
	typedef ArgSetBase<Arg,Base> Type;
    };
    
    template<typename Arg,typename Base>
    struct ArgSetRefIfRef<Arg&,Base>
    {
	typedef ArgSetBaseRef<Arg,Base> Type;
    };
    
    template<typename ArgSet,typename CAS>
    struct TransformArgSetHelper
    {
	typedef typename ArgSet::Arg Arg;
	typedef typename ArgSet::Base Base;
	
	typedef typename TransformArgSetResult<Arg,CAS>::Type NewArg;
	
	typedef TransformArgSetHelper<Base,CAS> BaseTransform;
	typedef typename BaseTransform::Type NewBase;

	typedef typename ArgSetRefIfRef<NewArg,NewBase>::Type Type;

	static Type doTransform(const ArgSet& as,const CAS& cas)
	{
	    return Type(transformArg(as.a,cas),BaseTransform::doTransform(as,cas));
	}
    };

    template<typename CAS>
    struct TransformArgSetHelper<ArgSet0,CAS>
    {
	typedef ArgSet0 Type;

	static Type doTransform(const ArgSet0& as,const CAS& cas)
	{
	    return as;
	}
    };
   
    template<typename ArgSet,typename CAS>
    typename TransformArgSetHelper<ArgSet,CAS>::Type transformArgSet(const ArgSet& as,const CAS& cas)
    {
	return TransformArgSetHelper<ArgSet,CAS>::doTransform(as,cas);
    }

    template<typename Arg>
    struct NumArgs
    {
	static const unsigned result=0;
    };
    
    template<unsigned i>
    struct NumArgs<FuncImpl::ParamType<i> >
    {
	static const unsigned result=i;
    };

    template<typename ArgSet>
    struct MinArgs;
    
    template<>
    struct MinArgs<ArgSet0>;
    

    template<typename ArgSet>
    struct MinArgs
    {
	static const unsigned result=CT::Max<unsigned,
					     NumArgs<typename ArgSet::Arg>::result,
					     MinArgs<typename ArgSet::Base>::result>::result;
    };
    
    template<>
    struct MinArgs<ArgSet0>
    {
	static const unsigned result=0;
    };
}


#endif
