#ifndef COMPOSER_ARGSETS_HH
#define COMPOSER_ARGSETS_HH

namespace ArgSets
{
    struct ArgSet0
    {};

    template<typename A,typename Next=ArgSet0>
    struct ArgSetBase:
	public Next
    {
	typedef A Arg;
	typedef Next Base;
	const Arg a;

	ArgSetBase(const Arg& a_,const Base& base_):
	    Base(base_),a(a_)
	{}
    };
    
    template<typename A>
    struct ArgSetBase<A,ArgSet0>:
	public ArgSet0
    {
	typedef A Arg;
	typedef ArgSet0 Base;
	const Arg a;
	
	ArgSetBase(const Arg& a_):
	    a(a_)
	{}

	ArgSetBase(const Arg& a_,const ArgSet0&):
	    a(a_)
	{}
    };

    template<typename A,typename Next=ArgSet0>
    struct ArgSetBaseRef:
	public Next
    {
	typedef A Arg;
	typedef Next Base;
	const Arg& a;

	ArgSetBaseRef(const Arg& a_,const Base& base_):
	    Base(base_),a(a_)
	{}
    };
    
    template<typename A>
    struct ArgSetBaseRef<A,ArgSet0>:
	public ArgSet0
    {
	typedef A Arg;
	typedef ArgSet0 Base;
	const Arg& a;
	
	ArgSetBaseRef(const Arg& a_):
	    a(a_)
	{}

	ArgSetBaseRef(const Arg& a_,const ArgSet0&):
	    a(a_)
	{}
    };

    // avoid ref-to-ref problems
    template<typename A,typename Next>
    struct ArgSetBaseRef<A&,Next>:
	public Next
    {
	typedef A Arg;
	typedef Next Base;
	const Arg& a;

	ArgSetBaseRef(const Arg& a_,const Base& base_):
	    Base(base_),a(a_)
	{}
    };

    template<typename A>
    struct ArgSetBaseRef<A&,ArgSet0>:
	public ArgSet0
    {
	typedef A Arg;
	typedef ArgSet0 Base;
	const Arg& a;
	
	ArgSetBaseRef(const Arg& a_):
	    a(a_)
	{}

	ArgSetBaseRef(const Arg& a_,const ArgSet0&):
	    a(a_)
	{}
    };

    

template<typename A1>
struct ArgSet1:
	public ArgSetBase<A1>
{
    typedef ArgSetBase<A1> ArgSet;
    
    ArgSet1(const A1& a1):
	ArgSet(a1)
    {}
};

template<typename A1,typename A2>
struct ArgSet2:
    public ArgSetBase<A1,ArgSet1<A2> >
{
    typedef ArgSetBase<A1,ArgSet1<A2> > ArgSet;

    ArgSet2(const A1& a1,const A2& a2):
	ArgSet(a1,ArgSet::Base(a2))
    {}
};

template<typename A1,typename A2,typename A3>
struct ArgSet3:
    public ArgSetBase<A1,ArgSet2<A2,A3> >
{
    typedef ArgSetBase<A1,ArgSet2<A2,A3> > ArgSet;

    ArgSet3(const A1& a1,const A2& a2,const A3& a3):
	ArgSet(a1,ArgSet::Base(a2,a3))
    {}
};

template<typename A1,typename A2,typename A3,typename A4>
struct ArgSet4:
    public ArgSetBase<A1,ArgSet3<A2,A3,A4> >
{
    typedef ArgSetBase<A1,ArgSet3<A2,A3,A4> > ArgSet;

    ArgSet4(const A1& a1,const A2& a2,const A3& a3,const A4& a4):
	ArgSet(a1,ArgSet::Base(a2,a3,a4))
    {}
};

template<typename A1,typename A2,typename A3,typename A4,typename A5>
struct ArgSet5:
    public ArgSetBase<A1,ArgSet4<A2,A3,A4,A5> >
{
    typedef ArgSetBase<A1,ArgSet4<A2,A3,A4,A5> > ArgSet;

    ArgSet5(const A1& a1,const A2& a2,const A3& a3,const A4& a4,const A5& a5):
	ArgSet(a1,ArgSet::Base(a2,a3,a4,a5))
    {}
};

template<typename A1,typename A2,typename A3,typename A4,typename A5,
	 typename A6>
struct ArgSet6:
    public ArgSetBase<A1,ArgSet5<A2,A3,A4,A5,A6> >
{
    typedef ArgSetBase<A1,ArgSet5<A2,A3,A4,A5,A6> > ArgSet;

    ArgSet6(const A1& a1,const A2& a2,const A3& a3,const A4& a4,const A5& a5,
	    const A6& a6):
	ArgSet(a1,ArgSet::Base(a2,a3,a4,a5,a6))
    {}
};

template<typename A1,typename A2,typename A3,typename A4,typename A5,
	 typename A6,typename A7>
struct ArgSet7:
    public ArgSetBase<A1,ArgSet6<A2,A3,A4,A5,A6,A7> >
{
    typedef ArgSetBase<A1,ArgSet6<A2,A3,A4,A5,A6,A7> > ArgSet;

    ArgSet7(const A1& a1,const A2& a2,const A3& a3,const A4& a4,const A5& a5,
	    const A6& a6,const A7& a7):
	ArgSet(a1,ArgSet::Base(a2,a3,a4,a5,a6,a7))
    {}
};

template<typename A1,typename A2,typename A3,typename A4,typename A5,
	 typename A6,typename A7,typename A8>
struct ArgSet8:
    public ArgSetBase<A1,ArgSet7<A2,A3,A4,A5,A6,A7,A8> >
{
    typedef ArgSetBase<A1,ArgSet7<A2,A3,A4,A5,A6,A7,A8> > ArgSet;

    ArgSet8(const A1& a1,const A2& a2,const A3& a3,const A4& a4,const A5& a5,
	    const A6& a6,const A7& a7,const A8& a8):
	ArgSet(a1,ArgSet::Base(a2,a3,a4,a5,a6,a7,a8))
    {}
};

template<typename A1,typename A2,typename A3,typename A4,typename A5,
	 typename A6,typename A7,typename A8,typename A9>
struct ArgSet9:
    public ArgSetBase<A1,ArgSet8<A2,A3,A4,A5,A6,A7,A8,A9> >
{
    typedef ArgSetBase<A1,ArgSet8<A2,A3,A4,A5,A6,A7,A8,A9> > ArgSet;

    ArgSet9(const A1& a1,const A2& a2,const A3& a3,const A4& a4,const A5& a5,
	    const A6& a6,const A7& a7,const A8& a8,const A9& a9):
	ArgSet(a1,ArgSet::Base(a2,a3,a4,a5,a6,a7,a8,a9))
    {}
};

template<typename A1,typename A2,typename A3,typename A4,typename A5,
	 typename A6,typename A7,typename A8,typename A9,typename A10>
struct ArgSet10:
    public ArgSetBase<A1,ArgSet9<A2,A3,A4,A5,A6,A7,A8,A9,A10> >
{
    typedef ArgSetBase<A1,ArgSet9<A2,A3,A4,A5,A6,A7,A8,A9,A10> > ArgSet;

    ArgSet10(const A1& a1,const A2& a2,const A3& a3,const A4& a4,const A5& a5,
	     const A6& a6,const A7& a7,const A8& a8,const A9& a9,const A10& a10):
	ArgSet(a1,ArgSet::Base(a2,a3,a4,a5,a6,a7,a8,a9,a10))
    {}
};

typedef ArgSet0 ArgSetRef0;

template<typename A1>
struct ArgSetRef1:
	public ArgSetBaseRef<A1>
{
    typedef ArgSetBaseRef<A1> ArgSet;
    
    ArgSetRef1(const A1& a1):
	ArgSet(a1)
    {}
};

template<typename A1,typename A2>
struct ArgSetRef2:
    public ArgSetBaseRef<A1,ArgSetRef1<A2> >
{
    typedef ArgSetBaseRef<A1,ArgSetRef1<A2> > ArgSet;

    ArgSetRef2(const A1& a1,const A2& a2):
	ArgSet(a1,ArgSet::Base(a2))
    {}
};

template<typename A1,typename A2,typename A3>
struct ArgSetRef3:
    public ArgSetBaseRef<A1,ArgSetRef2<A2,A3> >
{
    typedef ArgSetBaseRef<A1,ArgSetRef2<A2,A3> > ArgSet;

    ArgSetRef3(const A1& a1,const A2& a2,const A3& a3):
	ArgSet(a1,ArgSet::Base(a2,a3))
    {}
};

template<typename A1,typename A2,typename A3,typename A4>
struct ArgSetRef4:
    public ArgSetBaseRef<A1,ArgSetRef3<A2,A3,A4> >
{
    typedef ArgSetBaseRef<A1,ArgSetRef3<A2,A3,A4> > ArgSet;

    ArgSetRef4(const A1& a1,const A2& a2,const A3& a3,const A4& a4):
	ArgSet(a1,ArgSet::Base(a2,a3,a4))
    {}
};

template<typename A1,typename A2,typename A3,typename A4,typename A5>
struct ArgSetRef5:
    public ArgSetBaseRef<A1,ArgSetRef4<A2,A3,A4,A5> >
{
    typedef ArgSetBaseRef<A1,ArgSetRef4<A2,A3,A4,A5> > ArgSet;

    ArgSetRef5(const A1& a1,const A2& a2,const A3& a3,const A4& a4,const A5& a5):
	ArgSet(a1,ArgSet::Base(a2,a3,a4,a5))
    {}
};

template<typename A1,typename A2,typename A3,typename A4,typename A5,
	 typename A6>
struct ArgSetRef6:
    public ArgSetBaseRef<A1,ArgSetRef5<A2,A3,A4,A5,A6> >
{
    typedef ArgSetBaseRef<A1,ArgSetRef5<A2,A3,A4,A5,A6> > ArgSet;

    ArgSetRef6(const A1& a1,const A2& a2,const A3& a3,const A4& a4,const A5& a5,
	    const A6& a6):
	ArgSet(a1,ArgSet::Base(a2,a3,a4,a5,a6))
    {}
};

template<typename A1,typename A2,typename A3,typename A4,typename A5,
	 typename A6,typename A7>
struct ArgSetRef7:
    public ArgSetBaseRef<A1,ArgSetRef6<A2,A3,A4,A5,A6,A7> >
{
    typedef ArgSetBaseRef<A1,ArgSetRef6<A2,A3,A4,A5,A6,A7> > ArgSet;

    ArgSetRef7(const A1& a1,const A2& a2,const A3& a3,const A4& a4,const A5& a5,
	    const A6& a6,const A7& a7):
	ArgSet(a1,ArgSet::Base(a2,a3,a4,a5,a6,a7))
    {}
};

template<typename A1,typename A2,typename A3,typename A4,typename A5,
	 typename A6,typename A7,typename A8>
struct ArgSetRef8:
    public ArgSetBaseRef<A1,ArgSetRef7<A2,A3,A4,A5,A6,A7,A8> >
{
    typedef ArgSetBaseRef<A1,ArgSetRef7<A2,A3,A4,A5,A6,A7,A8> > ArgSet;

    ArgSetRef8(const A1& a1,const A2& a2,const A3& a3,const A4& a4,const A5& a5,
	    const A6& a6,const A7& a7,const A8& a8):
	ArgSet(a1,ArgSet::Base(a2,a3,a4,a5,a6,a7,a8))
    {}
};

template<typename A1,typename A2,typename A3,typename A4,typename A5,
	 typename A6,typename A7,typename A8,typename A9>
struct ArgSetRef9:
    public ArgSetBaseRef<A1,ArgSetRef8<A2,A3,A4,A5,A6,A7,A8,A9> >
{
    typedef ArgSetBaseRef<A1,ArgSetRef8<A2,A3,A4,A5,A6,A7,A8,A9> > ArgSet;

    ArgSetRef9(const A1& a1,const A2& a2,const A3& a3,const A4& a4,const A5& a5,
	    const A6& a6,const A7& a7,const A8& a8,const A9& a9):
	ArgSet(a1,ArgSet::Base(a2,a3,a4,a5,a6,a7,a8,a9))
    {}
};

template<typename A1,typename A2,typename A3,typename A4,typename A5,
	 typename A6,typename A7,typename A8,typename A9,typename A10>
struct ArgSetRef10:
    public ArgSetBaseRef<A1,ArgSetRef9<A2,A3,A4,A5,A6,A7,A8,A9,A10> >
{
    typedef ArgSetBaseRef<A1,ArgSetRef9<A2,A3,A4,A5,A6,A7,A8,A9,A10> > ArgSet;

    ArgSetRef10(const A1& a1,const A2& a2,const A3& a3,const A4& a4,const A5& a5,
	     const A6& a6,const A7& a7,const A8& a8,const A9& a9,const A10& a10):
	ArgSet(a1,ArgSet::Base(a2,a3,a4,a5,a6,a7,a8,a9,a10))
    {}
};

// for a general argset, get the type from the base
template<typename ArgSet,unsigned i>
struct GetASEntry
{
private:
    typedef typename ArgSet::Base Base;
    typedef typename ArgSet::Arg Arg;
    typedef GetASEntry<Base,i-1> BaseEntry;
public:
    typedef typename BaseEntry::Type Type;
    static Type& getVal(const ArgSet& as)
    {
	return BaseEntry::getVal(as);
    }
};

// the zeroeth argument is nonexistent
template<typename ArgSet>
struct GetASEntry<ArgSet,0>
{};

// and the first is straight forward
template<typename ArgSet>
struct GetASEntry<ArgSet,1>
{    
    typedef const typename ArgSet::Arg Type;
    static Type& getVal(const ArgSet& as)
    {
	
	return as.a;
    }
};

template<typename ArgSet>
struct CountArgs
{
    static const unsigned result=1+CountArgs<typename ArgSet::Base>::result;
};

template<>
struct CountArgs<ArgSet0>
{
    static const unsigned result=0;
};

}


#endif
