#ifndef OPERATORS_HH
#define OPERATORS_HH

#include "compose.hh"
#include "exprtype.hh"
#include "refholder.hh"

namespace FuncOpImpl
{
#define DECLARE_BIN_OP(name,op)\
struct OpBin##name\
{\
    template<typename T,typename U>\
    struct ResultHelper\
    {\
	typedef typename ExprType::RValueLValueType<\
            typename EXPR_TYPE((ExprType::makeT<const T&>() op ExprType::makeT<const U&>()))\
            >::Type Type;\
    };\
    \
    template<typename T,typename U>\
    typename ResultHelper<T,U>::Type\
    operator()(const T& t,const U& u) const\
    {\
	return t op u;\
    }\
}
    
    DECLARE_BIN_OP(Plus,+);
    DECLARE_BIN_OP(Minus,-);
    DECLARE_BIN_OP(Multiply,*);
    DECLARE_BIN_OP(Divide,/);
    DECLARE_BIN_OP(Mod,%);
    DECLARE_BIN_OP(ShiftLeft,<<);
    DECLARE_BIN_OP(ShiftRight,>>);
    DECLARE_BIN_OP(LessThan,<);
    DECLARE_BIN_OP(GreaterThan,>);
    DECLARE_BIN_OP(LessEqual,<=);
    DECLARE_BIN_OP(GreaterEqual,>=);
    DECLARE_BIN_OP(Equal,==);
    DECLARE_BIN_OP(NotEqual,!=);
    DECLARE_BIN_OP(BitAnd,&);
    DECLARE_BIN_OP(BitOr,|);
    DECLARE_BIN_OP(BitXor,^);
    DECLARE_BIN_OP(LogicalAnd,&&);
    DECLARE_BIN_OP(LogicalOr,||);
}

namespace FuncImpl
{
#define SPECIALIZE_HASRESULTHELPER_OP(name)\
    template<>\
    struct HasResultHelper<FuncOpImpl::Op##name>\
    {\
	static const bool result=true;\
    }
    
    SPECIALIZE_HASRESULTHELPER_OP(BinPlus);
    SPECIALIZE_HASRESULTHELPER_OP(BinMinus);
    SPECIALIZE_HASRESULTHELPER_OP(BinMultiply);
    SPECIALIZE_HASRESULTHELPER_OP(BinDivide);
    SPECIALIZE_HASRESULTHELPER_OP(BinMod);
    SPECIALIZE_HASRESULTHELPER_OP(BinShiftLeft);
    SPECIALIZE_HASRESULTHELPER_OP(BinShiftRight);
    SPECIALIZE_HASRESULTHELPER_OP(BinLessThan);
    SPECIALIZE_HASRESULTHELPER_OP(BinGreaterThan);
    SPECIALIZE_HASRESULTHELPER_OP(BinLessEqual);
    SPECIALIZE_HASRESULTHELPER_OP(BinGreaterEqual);
    SPECIALIZE_HASRESULTHELPER_OP(BinEqual);
    SPECIALIZE_HASRESULTHELPER_OP(BinNotEqual);
    SPECIALIZE_HASRESULTHELPER_OP(BinBitAnd);
    SPECIALIZE_HASRESULTHELPER_OP(BinBitOr);
    SPECIALIZE_HASRESULTHELPER_OP(BinBitXor);
    SPECIALIZE_HASRESULTHELPER_OP(BinLogicalAnd);
    SPECIALIZE_HASRESULTHELPER_OP(BinLogicalOr);

    // specialize CallerHelper for the logical operators, to enable short-circuit evaluation
    template<typename ArgSet,typename CAS>
    struct CallerHelper<FuncOpImpl::OpBinLogicalAnd,ArgSet,CAS,2>
    {
	typedef typename ArgSets::TransformArgSetHelper<ArgSet,CAS>::Type TransformedArgSet;
	typedef typename RTHelper<FuncOpImpl::OpBinLogicalAnd,TransformedArgSet>::Type Type;

	static Type callFunc(const FuncOpImpl::OpBinLogicalAnd&,const ArgSet& as,const CAS& cas)
	{
	    typedef ArgSets::GetASEntry<ArgSet,1> Arg1;
	    typedef ArgSets::GetASEntry<ArgSet,2> Arg2;

	    return transformArg(Arg1::getVal(as),cas) && transformArg(Arg2::getVal(as),cas);
        }
    };

    template<typename ArgSet,typename CAS>
    struct CallerHelper<FuncOpImpl::OpBinLogicalOr,ArgSet,CAS,2>
    {
	typedef typename ArgSets::TransformArgSetHelper<ArgSet,CAS>::Type TransformedArgSet;
	typedef typename RTHelper<FuncOpImpl::OpBinLogicalOr,TransformedArgSet>::Type Type;

	static Type callFunc(const FuncOpImpl::OpBinLogicalOr&,const ArgSet& as,const CAS& cas)
	{
	    typedef ArgSets::GetASEntry<ArgSet,1> Arg1;
	    typedef ArgSets::GetASEntry<ArgSet,2> Arg2;

	    return transformArg(Arg1::getVal(as),cas) || transformArg(Arg2::getVal(as),cas);
        }
    };

}


namespace FuncOpImpl
{
    
#define DECLARE_UN_OP(name,op)\
struct OpUn##name\
{\
    template<typename T>\
    struct ResultHelper\
    {\
	typedef typename ExprType::RValueLValueType<\
	    typename EXPR_TYPE(op ExprType::makeT<const T&>())\
            >::Type Type;\
    };\
    \
    template<typename T>\
    typename ResultHelper<T>::Type\
    operator()(const T& t) const\
    {\
	return op t;\
    }\
}

#define DECLARE_UN_POST_OP(name,op)\
struct OpUn##name\
{\
    template<typename T>\
    struct ResultHelper\
    {\
	typedef typename ExprType::RValueLValueType<\
	    typename EXPR_TYPE(ExprType::makeT<const T&>() op)\
            >::Type Type;\
    };\
    \
    template<typename T>\
    typename ResultHelper<T>::Type\
    operator()(const T& t) const\
    {\
	return t op;\
    }\
};

DECLARE_UN_OP(PreInc,++);
DECLARE_UN_OP(PreDec,--);
DECLARE_UN_OP(AddressOf,&);
DECLARE_UN_OP(DeRef,*);
DECLARE_UN_OP(Not,!);
DECLARE_UN_OP(Invert,~);
DECLARE_UN_OP(Minus,-);
DECLARE_UN_POST_OP(PostInc,++);
DECLARE_UN_POST_OP(PostDec,--);

}
namespace FuncImpl
{
SPECIALIZE_HASRESULTHELPER_OP(UnPreInc);
SPECIALIZE_HASRESULTHELPER_OP(UnPreDec);
SPECIALIZE_HASRESULTHELPER_OP(UnAddressOf);
SPECIALIZE_HASRESULTHELPER_OP(UnDeRef);
SPECIALIZE_HASRESULTHELPER_OP(UnNot);
SPECIALIZE_HASRESULTHELPER_OP(UnInvert);
SPECIALIZE_HASRESULTHELPER_OP(UnMinus);
SPECIALIZE_HASRESULTHELPER_OP(UnPostInc);
SPECIALIZE_HASRESULTHELPER_OP(UnPostDec);
}


namespace FuncImpl
{
#define FORWARD_BIN_OP(name,op)    \
template<typename T,typename U>\
FuncImpl::ComposedFunc<FuncOpImpl::OpBin##name,ArgSets::ArgSet2<T,U> > operator op(const T& t,const U& u)\
{\
    return Compose::compose(FuncOpImpl::OpBin##name(),t,u);\
}
    FORWARD_BIN_OP(Plus,+)
    FORWARD_BIN_OP(Minus,-)
    FORWARD_BIN_OP(Multiply,*)
    FORWARD_BIN_OP(Divide,/)
    FORWARD_BIN_OP(Mod,%)
    FORWARD_BIN_OP(ShiftLeft,<<)
    FORWARD_BIN_OP(ShiftRight,>>)
    FORWARD_BIN_OP(LessThan,<)
    FORWARD_BIN_OP(GreaterThan,>)
    FORWARD_BIN_OP(LessEqual,<=)
    FORWARD_BIN_OP(GreaterEqual,>=)
    FORWARD_BIN_OP(Equal,==)
    FORWARD_BIN_OP(NotEqual,!=)
    FORWARD_BIN_OP(BitAnd,&)
    FORWARD_BIN_OP(BitOr,|)
    FORWARD_BIN_OP(BitXor,^)
    FORWARD_BIN_OP(LogicalAnd,&&)
    FORWARD_BIN_OP(LogicalOr,||)

#define FORWARD_UN_OP(name,op)    \
template<typename T>\
FuncImpl::ComposedFunc<FuncOpImpl::OpUn##name,ArgSets::ArgSet1<T> > operator op(const T& t)\
{\
    return Compose::compose(FuncOpImpl::OpUn##name(),t);\
}

#define FORWARD_UN_POST_OP(name,op)    \
template<typename T>\
FuncImpl::ComposedFunc<FuncOpImpl::OpUn##name,ArgSets::ArgSet1<T> > operator op(const T& t,int)\
{\
    return Compose::compose(FuncOpImpl::OpUn##name(),t);\
}

FORWARD_UN_OP(PreInc,++)
FORWARD_UN_OP(PreDec,--)
FORWARD_UN_OP(AddressOf,&)
FORWARD_UN_OP(DeRef,*)
FORWARD_UN_OP(Not,!)
FORWARD_UN_OP(Invert,~)
FORWARD_UN_OP(Minus,-)
FORWARD_UN_POST_OP(PostInc,++)
FORWARD_UN_POST_OP(PostDec,--)

template<typename T>
struct DoCastTo
{
    typedef T result_type;
    template<typename U>
    T operator()(const U& u) const
    {
	return T(u);
    }
};

}

namespace Compose
{
    template<typename T,typename U>
    FuncImpl::ComposedFunc<FuncImpl::DoCastTo<T>,ArgSets::ArgSet1<U> > castTo(const U& u,T*p=0)
    {
	return compose(FuncImpl::DoCastTo<T>(),u);
    }

//     template<typename T,typename U>
//     FuncImpl::ComposedFunc<FuncImpl::DoCastTo<T&>,ArgSets::ArgSet1<U> > castToRef(const U& u,T*p=0)
//     {
// 	return compose(FuncImpl::DoCastTo<T&>(),u);
//     }
}

// special cases of operators
namespace FuncOpImpl
{
    // dereferencing a pointer always yields a reference to the pointed-to type
    // no need for EXPR_TYPE here.
    template<typename T>
    struct OpUnDeRef::ResultHelper<T*>
    {
	typedef T& Type;
    };
    
    template<typename T>
    struct OpUnDeRef::ResultHelper<T* const>
    {
	typedef T& Type;
    };

    template<typename T>
    struct OpUnDeRef::ResultHelper<T* volatile>
    {
	typedef T& Type;
    };

    template<typename T>
    struct OpUnDeRef::ResultHelper<T* const volatile>
    {
	typedef T& Type;
    };
}


#endif
