#ifndef REFHOLDER_HH
#define REFHOLDER_HH

#include "exprtype.hh"

namespace Holder
{
    // for passing by reference
    template<typename T>
    class RefHolder
    {
    public:
	typedef T ContainedType;
	
	RefHolder(T& ref_):
	    ref(ref_)
	{}
	operator T&() const
	{
	    return ref;
	}
	T* operator&() const
	{
	    return &ref;
	}
    private:
	T& ref;
    };

    template<typename T>
    struct RefType
    {
	typedef RefHolder<T> Type;
    };
    
    template<typename T>
    struct RefType<RefHolder<T> >
    {
	typedef RefHolder<T> Type;
    };

    template<typename T>
    struct RefType<const RefHolder<T> >
    {
	typedef const RefHolder<T> Type;
    };
}
namespace Compose
{
    template<typename T>
    const Holder::RefHolder<T> byRef(T& ref)
    {
	return ref;
    }

    template<typename T>
    const Holder::RefHolder<const T> byRef(const T& ref)
    {
	return ref;
    }

    template<typename T>
    const Holder::RefHolder<T> byRef(Holder::RefHolder<T>& ref)
    {
	return ref;
    }

    template<typename T>
    const Holder::RefHolder<T> byRef(const Holder::RefHolder<T>& ref)
    {
	return ref;
    }
}

#endif
