#include <iostream>
#include <typeinfo>
#include <string>
#include <algorithm>
#include <sstream>
#include <vector>
#include "lambda.hh"
#include <time.h>

using namespace Compose;

REGISTER_EXPR_TYPE(std::string,100)
REGISTER_EXPR_TYPE(Holder::RefHolder<int>,101)
REGISTER_EXPR_TYPE(Holder::RefHolder<double>,102)
REGISTER_EXPR_TYPE(Holder::RefHolder<std::ostream>,104)
REGISTER_EXPR_TYPE(std::ostream,105)

    struct F
    {
	typedef int result_type;
	int operator()(int i) const
	{
	    return i;
	}
    } f;

struct Write
{
    typedef void result_type;
    template<typename T>
    void operator()(const T& i) const
    {
	std::cout<<"i="<<i<<std::endl;
    }
}
write;

namespace FuncImpl
{
    template<>
    struct HasResultType<Write>
    {
	static const bool result=true;
    };
}

struct Write2
{
    typedef std::ostream& result_type;
    template<typename T1>
    std::ostream& operator()(std::ostream& os,const T1& t1) const
    {
	return os<<t1<<std::endl;
    }
    template<typename T1,typename T2>
    std::ostream& operator()(std::ostream& os,const T1& t1,const T2& t2) const
    {
	return os<<t1<<t2<<std::endl;
    }
}
write2;

int increment(int& i)
{
    return ++i;
};

struct DoIncrement
{
    int operator()(int& i) const
    {
	return ++i;
    }
} doInc;


struct OpUnIncrement
{
    template<typename T>
    struct OpUnHelper
    {
	typedef typename EXPR_TYPE(++ExprType::makeT<const T&>()) Type;
    };
    
    template<typename T>
    typename OpUnHelper<T>::Type operator()(const T& t) const
    {
	return ++t;
    }
} opUnIncrement;

	struct in_range { 
	    int imin, imax; 
	    in_range(int x, int y) : imin(x), imax(y) {} 
	    bool operator()(int i) const { return imin <= i && i < imax; } 
	}; 

	struct do_math { 
	    int imin, imax; 
	    do_math(int x, int y) : imin(x), imax(y) {} 
	    int operator()(int i) const { return (imin *i) + (i % imax); } 
	}; 

bool vv[100]; 
int mr[100];

struct X
{};

std::ostream& operator<<(std::ostream& os,const X& x)
{
    os<<"Writing X object";
    return os;
}

bool shortCircuitTrace(bool x)
{
    std::cout<<"Trace bool="<<x<<std::endl;
}


int main()
{
//     byRef<std::ostream>(std::cout)<<"hello"<<std::endl;
    

    EXPR_TYPE(3) i=3;
    std::cout<<typeid(i).name()<<std::endl;
    
    EXPR_TYPE(i+3) j=i+3;
    std::cout<<typeid(j).name()<<std::endl;
    EXPR_TYPE(i+3.5) k=i+3.5;
    std::cout<<typeid(k).name()<<std::endl;
    std::cout<<((compose(f,3)+1)())<<std::endl;
    std::cout<<((compose(f,_1)+2.9)(1.5))<<std::endl;
    std::cout<<(7.5-(compose(f,_1))(1.2))<<std::endl;
    std::cout<<typeid(7.5-(compose(f,_1))).name()<<std::endl;
    std::cout<<(_1+3)(1)<<std::endl;
    std::cout<<(_1+3)(1.5)<<std::endl;
    std::cout<<typeid((_1+3)(1)).name()<<std::endl;
    std::cout<<(_1+(const char*)("world"))(std::string("hello, "))<<std::endl;
    
    int x=3;
    std::cout<<"x="<<x<<std::endl;
    std::cout<<"++x="<<compose(increment,_1)(byRef(x))<<std::endl;
    std::cout<<"++x="<<compose(opUnIncrement,_1)(byRef(x))<<std::endl;
    std::cout<<"++x="<<(++_1)(byRef(x))<<std::endl;
    std::cout<<"x++="<<(_1++)(byRef(x))<<std::endl;
    std::cout<<"x="<<x<<std::endl;
    std::cout<<"--x="<<(--_1)(byRef(x))<<std::endl;
    std::cout<<"x="<<x<<std::endl;

    X xVar;
    X* px=&xVar;
    
    std::cout<<"deref test for custom class: *px="<<(*_1)(px)<<std::endl;
    
    
    std::cout<<"address-of test: &x="<<(&_1)(byRef(x))<<std::endl;
    std::cout<<"real &x="<<(&x)<<std::endl;
    std::cout<<"deref test: x="<<(*_1)(&x)<<std::endl;

    std::cout<<"Invert test: ~x="<<(~_1)(x)<<std::endl;
    std::cout<<"Not test: !x="<<(!_1)(x)<<std::endl;
    std::cout<<"Not test: !0="<<(!_1)(0)<<std::endl;
    std::cout<<"Neg test: -x="<<(-_1)(x)<<std::endl;
    std::cout<<"x="<<x<<std::endl;

    std::cout<<"Shift left test: x<<2="<<(_1<<2)(x)<<std::endl;
    std::cout<<"Shift left ref test: x<<2="<<(byRef(_1)<<2)(x)<<std::endl;

    std::cout<<"x<=10?"<<(_1<=10)(x)<<std::endl;
    std::cout<<"x>10?"<<(_1>10)(x)<<std::endl;


    std::vector<int> vec;
    {
	
	for(unsigned i=0;i<10;++i)
	{
	    vec.push_back(i);
	}
    }

    std::for_each(vec.begin(),vec.end(),compose(write,_1+5*_1*_1));

    

    compose(write,_1)(1.4);

    // for_each and  simple compose
    std::for_each(vec.begin(),vec.end(),compose(write,_1));

    // for_each and compose, operator+
    std::for_each(vec.begin(),vec.end(),compose(write,_1+2.5));

    std::for_each(vec.begin(),vec.end(),compose(write,_1<<1));
    std::for_each(vec.begin(),vec.end(),compose(write,byRef(x)<<_1));

    //  cannot pass non-const ref, as we need to use byRef(), which we can't do    
    // for_each and compose, operator++
    std::for_each(vec.begin(),vec.end(),++byRef(_1));

    // for_each and compose, operator*, byRef
    std::for_each(vec.begin(),vec.end(),compose(voidReturn(write2),byRef<std::ostream>(std::cout),_1*_1));

    std::ostringstream oss;

    std::for_each(vec.begin(),vec.end(),compose(voidReturn(write2),byRef<std::ostream>(oss),(const char*)("_1*_1="),_1*_1));
    std::cout<<"oss.str()="<<oss.str();
    std::for_each(vec.begin(),vec.end(),compose(voidReturn(write2),byRef<std::ostream>(std::cout),(const char*)("_1*_1="),_1*_1));
    std::for_each(vec.begin(),vec.end(),compose(voidReturn(write2),byRef<std::ostream>(std::cout),_1*_1));
    std::for_each(vec.begin(),vec.end(),compose(voidReturn(write2),byRef<std::ostream>(std::cout),_1*_1,(const char*)("=_1*_1")));

    
    std::cout<<"1<<3="<<(1<<_1)(3)<<std::endl;

    // for_each
    std::for_each(vec.begin(),vec.end(),compose(write,_1));

    std::cout<<"Testing short-circuit evaluation"<<std::endl;
    std::cout<<"false && false"<<std::endl;
    (compose(shortCircuitTrace,_1) && compose(shortCircuitTrace,_2))(false,false);
    std::cout<<"false && true"<<std::endl;
    (compose(shortCircuitTrace,_1) && compose(shortCircuitTrace,_2))(false,true);
    std::cout<<"true && false"<<std::endl;
    (compose(shortCircuitTrace,_1) && compose(shortCircuitTrace,_2))(true,false);
    std::cout<<"true && true"<<std::endl;
    (compose(shortCircuitTrace,_1) && compose(shortCircuitTrace,_2))(true,true);

    std::cout<<"false || false"<<std::endl;
    (compose(shortCircuitTrace,_1) || compose(shortCircuitTrace,_2))(false,false);
    std::cout<<"false || true"<<std::endl;
    (compose(shortCircuitTrace,_1) || compose(shortCircuitTrace,_2))(false,true);
    std::cout<<"true || false"<<std::endl;
    (compose(shortCircuitTrace,_1) || compose(shortCircuitTrace,_2))(true,false);
    std::cout<<"true || true"<<std::endl;
    (compose(shortCircuitTrace,_1) || compose(shortCircuitTrace,_2))(true,true);
    
    
    {
	
    int v[100] = {0,1,2,3,4,5,6,7,8,9,
		  10,11,12,13,14,15,16,17,18,19,
		  20,21,22,23,24,25,26,27,28,29,
		  30,31,32,33,34,35,36,37,38,39,
		  40,41,42,43,44,45,46,47,48,49,
		  50,51,52,53,54,55,56,57,58,59,
		  60,61,62,63,64,65,66,67,68,69,
		  70,71,72,73,74,75,76,77,78,79,
		  80,81,82,83,84,85,86,87,88,89,
		  90,91,92,93,94,95,96,97,98,99
    };
    
	const int r1 = 30, r2 = 50; 
	const unsigned loopCount=1000000;
	time_t now;
	time_t end;

	std::cout<<"small array, many times"<<std::endl;

	// using lambda: 
	now=time(NULL);
	
	{
	    for(unsigned i=0;i<loopCount;++i)
		std::transform(v, &v[100],vv, r1 <= _1 && _1 < r2); 
	}
	end=time(NULL);
	std::cout<<"ET logical expr: "<<end-now<<std::endl;
	
	// using classic STL: 
	now=time(NULL);
	{
	    for(unsigned i=0;i<loopCount;++i)
		std::transform(v, &v[100],vv, in_range(r1, r2)); 
	}
	end=time(NULL);
	std::cout<<"Classic logical expr: "<<end-now<<std::endl;
	
	std::vector<int> vec,vec2;
	vec.reserve(loopCount);
	vec2.resize(loopCount);
	{
	    for(unsigned i=0;i<loopCount;++i)
	    {
		vec.push_back(i%100);
	    }
	}

	// using lambda: 
	now=time(NULL);
	
	{
	    for(unsigned i=0;i<loopCount;++i)
		std::transform(v, &v[100],mr, (r1 *_1) + (_1 % r2)); 
	}
	end=time(NULL);
	std::cout<<"ET math: "<<end-now<<std::endl;
	
	// using classic STL: 
	now=time(NULL);
	{
	    for(unsigned i=0;i<loopCount;++i)
		std::transform(v, &v[100],mr, do_math(r1, r2)); 
	}
	end=time(NULL);
	std::cout<<"Classic math: "<<end-now<<std::endl;
	
	std::cout<<"large array, few times"<<std::endl;
	
	// using lambda: 
	now=time(NULL);
	
	{
	    for(unsigned i=0;i<100;++i)
		std::transform(vec.begin(), vec.end(),vec2.begin(), r1 <= _1 && _1 < r2); 
	}
	end=time(NULL);
	std::cout<<"ET logical expr: "<<end-now<<std::endl;
	
	// using classic STL: 
	now=time(NULL);
	{
	    for(unsigned i=0;i<100;++i)
		std::transform(vec.begin(), vec.end(),vec2.begin(), in_range(r1, r2)); 
	}
	end=time(NULL);
	std::cout<<"Classic logical expr: "<<end-now<<std::endl;
	

	// using lambda: 
	now=time(NULL);
	
	{
	    for(unsigned i=0;i<100;++i)
		std::transform(vec.begin(), vec.end(),vec2.begin(), (r1 *_1) + (_1 % r2)); 
	}
	end=time(NULL);
	std::cout<<"ET math: "<<end-now<<std::endl;
	
	// using classic STL: 
	now=time(NULL);
	{
	    for(unsigned i=0;i<100;++i)
		std::transform(vec.begin(), vec.end(),vec2.begin(), do_math(r1, r2)); 
	}
	end=time(NULL);
	std::cout<<"Classic math: "<<end-now<<std::endl;
	

	std::cout<<"back to small array"<<std::endl;
	// using lambda, preinc
	now=time(NULL);
	
	{
	    for(unsigned i=0;i<loopCount;++i)
		std::for_each(v, &v[100],++byRef(_1)); 
	}
	end=time(NULL);
	std::cout<<"ET preinc: "<<end-now<<std::endl;
	
	// using classic STL: 
	now=time(NULL);
	{
	    for(unsigned i=0;i<loopCount;++i)
		std::for_each(v, &v[100], increment); 
	}
	end=time(NULL);
	std::cout<<"Classic inc with function: "<<end-now<<std::endl;
	
	// using classic STL and function object: 
	now=time(NULL);
	{
	    for(unsigned i=0;i<loopCount;++i)
		std::for_each(v, &v[100], doInc); 
	}
	end=time(NULL);
	std::cout<<"Classic inc with function object: "<<end-now<<std::endl;
    }
    
	return 0;
	
}

